import numpy as np
import scipy.sparse as sp

import pickle
import scipy

import __builtin__

try:
    __builtin__.profile
except AttributeError:
    # No line profiler, provide a pass-through version
    def profile(func): return func
    __builtin__.profile = profile

"""
authors: Clement Gehring
contact: gehring@csail.mit.edu
date: May 2015
"""

class CompressedMatrix(object):
    """ Constructor for a finite rank approximation of a square matrix.
    
        max_rank: maximum allowable rank for the approximation
        
        size: size of the square matrix (i.e., matrix is a size x size)
        
        deferred_updates:   boolean to tell whether updates should always be
                            rank one or if they should be deferred. Currently,
                            when deferred, updates are always of rank 
                            'max_rank'. This could be modified in the future.
    """
    def __init__(self, 
                 max_rank, 
                 size,
                 deferred_updates):
        
        # SVD matrices for the small rank approximation
        self.matrices = (np.zeros((size, 1)),
                         np.zeros(1),
                         np.zeros((size, 1)))
        
        # flag to know whether updates are deferred
        self.buffered = deferred_updates
        
        # allocate buffer matrices, if need be
        if self.buffered:
            self.ab_buffer = (np.zeros((size, max_rank)),
                              np.zeros((size, max_rank)))
            
        # flag to know whether the matrix has been initialized with some
        # non-zero quantities
        self.initialized = False
        
        # variable to count the number of buffered updates, not used
        # if we don't defer the updates.
        self.count = 0
        
        # maximum rank of the approximation
        self.max_rank = max_rank
        
        # holder for the L and R matrices for optimizing the rank one updates
        self.tmpmatrices = None
        
    """ Add an vector outer product ab^T to the compressed matrix.
    """
    @profile
    def update_matrix(self, a, b):
        if self.buffered:
            if a.ndim == 1:
                a = a.reshape((-1,1))
            if b.ndim == 1:
                b = b.reshape((-1,1))
                
            # save the current update
            A, B = self.ab_buffer
            
            # if the whole update doesn't fit in the buffer, iterate over
            # chunks that do
            j = 0
            while j<a.shape[1]:
                i = self.count + a.shape[1] - j
                i_clamp = min(i, self.max_rank)
                jp = j + i_clamp - self.count
                
                # place sub-matrix into buffer, special case if sparse....
                mat_a = a[:, j:jp]
                if sp.issparse(mat_a):
                    I, J = mat_a.nonzero()
                    A[I, J+self.count] = mat_a.data
                else:
                    A[:, self.count:i_clamp] = mat_a
                   
                # place sub-matrix into buffer, special case if sparse.... 
                mat_b = b[:, j:jp]
                if sp.issparse(mat_b):
                    I, J = mat_b.nonzero()
                    B[I, J+self.count] = mat_b.data
                else:
                    B[:, self.count:i_clamp] = mat_b
                
                self.count = i_clamp
                j = jp
                
                # if the buffer is full, update svd
                if self.count >= self.max_rank:
                    self.matrices = self.update_and_clear_buffer()
        else:
            if a.ndim > 1:
                for i in xrange(a.shape[1]):
                    vec_a = a[:,i]
                    vec_b = b[:,i]
                    self.matrices = self.optimized_rank1_update_svd(vec_a, vec_b)
            else:
                self.matrices = self.optimized_rank1_update_svd(a, b)
            
    """ Update svd with a rank-k outer-product, where k = self.count
    """           
    def update_and_clear_buffer(self):
        A, B = self.ab_buffer
        matrices = self.rankk_update_svd((A[:,:self.count], B[:,:self.count]))
        A[:,:] = 0.0
        B[:,:] = 0.0
        self.count = 0
        return matrices
        
    """ Incremental SVD update for some rank>1 update. In principal, this
        should not be called from outside this class (though nothing bad
        is going to happen if you do).
    """
    @profile
    def rankk_update_svd(self, ab_buffer):
        U,S,V = self.matrices
        A, B = ab_buffer
        Up = U.copy()
        Vp = V.copy()
        Sp = S.copy()
        if self.initialized:
            # compute value for the left and right subspace
            Q_a, R_a = np.linalg.qr(A - U.dot(U.T.dot(A)), mode='reduced')
            Q_b, R_b = np.linalg.qr(B - V.dot(V.T.dot(B)), mode='reduced')

            Ap = np.vstack((U.T.dot(A), R_a))
            Bp = np.vstack((V.T.dot(B), R_b))

            success = 0
            try:
                # naive diagonalization of the center matrix (i.e., with SVD)
                K = np.diag(np.hstack((S, np.zeros(R_a.shape[0])))) + Ap.dot(Bp.T)
                Up, Sp, Vp = np.linalg.svd(K, full_matrices = False)
                success = 1
            except:    
                print "WARNING: SVD failed to converge, this shouldn't happen, continuing anyway ...\n"

            if success == 1:
                # update left and right singular vectors
                U = np.hstack((U, Q_a)).dot(Up)
                V = np.hstack((V, Q_b)).dot(Vp.T)
            
                self.rank = min(self.max_rank, Sp.size)
            
        else:
            # initialize, assuming matrix is all zeroes
            
            # compute value for the left and right subspace
            Q_a, R_a = np.linalg.qr(A, mode='reduced')
            Q_b, R_b = np.linalg.qr(B, mode='reduced')
            
            # initial diagonalization 
            Up, Sp, Vp = np.linalg.svd(R_a.dot(R_b.T), full_matrices = False)

            # construct the singular vectors
            U = Q_a.dot(Up)
            V = Q_b.dot(Vp.T)
            self.initialized = True
            
            # Currently, we don't truncate singular values close to zero and
            # keep the full max_rank approximation. We could potentially remove
            # singular values close to zero.
            self.rank = min(self.max_rank, Sp.size)
            
        S = Sp[:self.rank]
        U = U[:,:self.rank]
        V = V[:,:self.rank]
        return (U, S, V)
    
    """ Incremental SVD update for some rank=1 update. In principal, this
        should not be called from outside this class. 
    """
    @profile
    def rank1_update_svd(self, a, b):
        U,S,V = self.matrices
        
        if a.ndim == 1:
            a = a.reshape((-1,1))
        if b.ndim == 1:
            b = b.reshape((-1,1))
            
        if self.initialized:
            # compute value for the left subspace
            m = (a.T.dot(U)).T
            p = a - U.dot(m)
            r_a = np.linalg.norm(p)
            
            # avoid NaNs
            if r_a < 1e-6:
                p[:] = 0.0
                r_a = 0.0
                P = np.zeros_like(p)
            else:
                P = p/r_a
            
            # compute values for the right subspace
            n = (b.T.dot(V)).T
            q = b - V.dot(n)
            r_b = np.linalg.norm(q)
            
            # avoid NaNs
            if r_b < 1e-6:
                q[:] = 0.0
                r_b = 0.0
                Q = np.zeros_like(q)
            else:
                Q = q/r_b

            # naive diagonalization of the center matrix (i.e., with SVD)
            K = np.diag(np.hstack((S, np.zeros(1)))) \
                    + np.outer( np.vstack((m, r_a)), np.vstack((n, r_b)))
            
            Up, Sp, Vp = np.linalg.svd(K, full_matrices = False)

            # update left and right singular vectors
            U = np.hstack((U, P)).dot(Up)
            V = np.hstack((V, Q)).dot(Vp.T)
            
            # increase the current rank approximation (capped at max_rank):
            
            # this could be modified to not allow rank increases if the
            # last singular value is very close to zero. For now it always
            # increases as long as it has not reached the max.
            
            self.rank = min(self.rank + 1, self.max_rank)

        else:
            # initialize, assuming matrix is all zeroes
            if sp.issparse(a):
                a = a.toarray()
                
            if sp.issparse(b):
                b = b.toarray()

            r_a = np.linalg.norm(a)
            r_b = np.linalg.norm(b)
            
            Sp = np.array([r_a*r_b])

            U = (a/r_a).reshape((-1,1))
            V = (b/r_b).reshape((-1,1))
            self.initialized = True
            self.rank = 1
            
        S = Sp[:self.rank]
        U = U[:,0:self.rank]
        V = V[:,0:self.rank]
        return (U, S, V)

    """ Incremental SVD update for some rank=1 update. In principal, this
        should not be called from outside this class. This method has a 
        few minor optimizations.
    """

    @profile
    def optimized_rank1_update_svd(self, a, b):    
        U,S,V = self.matrices
        
        
        if a.ndim == 1:
            a = a.reshape((-1,1))
        if b.ndim == 1:
            b = b.reshape((-1,1))
            
        if self.initialized:
            L,R = self.tmpmatrices
            
            # compute value for the left subspace
            m = (a.T.dot(U)).T
            p = a - U.dot(m)
            r_a = np.linalg.norm(p)
            
            # avoid NaNs
            if r_a < 1e-6:
                p[:] = 0.0
                r_a = 0.0
                P = np.zeros_like(p)
            else:
                P = p/r_a
            
            # compute values for the right subspace
            n = (b.T.dot(V)).T
            q = b - V.dot(n)
            r_b = np.linalg.norm(q)
            
            # avoid NaNs
            if r_b < 1e-6:
                q[:] = 0.0
                r_b = 0.0
                Q = np.zeros_like(q)
            else:
                Q = q/r_b

            # naive diagonalization of the center matrix (i.e., with SVD)
            K = np.diag(np.hstack((S, np.zeros(1)))) \
                    + np.outer( np.vstack((m, r_a)), np.vstack((n, r_b)))
            
            Up, Sp, Vp = np.linalg.svd(K, full_matrices = False)

            # update left and right singular vectors
            L = scipy.linalg.block_diag(L, np.ones((1,1))).dot(Up)
            R = scipy.linalg.block_diag(R, np.ones((1,1))).dot(Vp.T)
            U = np.hstack((U, P))
            V = np.hstack((V, Q))
            self.rank += 1
            
            if self.rank >= self.rank*2 :
                U = U.dot(L)
                V = V.dot(R)
                self.rank = self.max_rank
                L = np.eye(self.rank)
                R = np.eye(self.rank)
            

        else:
            # initialize, assuming matrix is all zeroes
            if sp.issparse(a):
                a = a.toarray()
                
            if sp.issparse(b):
                b = b.toarray()

            r_a = np.linalg.norm(a)
            r_b = np.linalg.norm(b)
            
            Sp = np.array([r_a*r_b])

            U = (a/r_a).reshape((-1,1))
            V = (b/r_b).reshape((-1,1))
            self.initialized = True
            self.rank = 1
            L = np.eye(self.rank)
            R = np.eye(self.rank)
            
        S = Sp[:self.rank]
        U = U[:,0:self.rank]
        V = V[:,0:self.rank]
        self.tmpmatrices = (L, R)
        return (U, S, V)
    
    """ Compute the matrix-vector Mx product without constructing the full
        matrix, where M is the compressed matrix and x is an input vector.
    """
    def mat_vec_product(self, x):
        U,S,V = self.matrices
        
        # this term is written this way such that, if x is a sparse vector,
        # x's dot method is called. (Its clunky but its the only way I know)
        xV = x.T.dot(V)
        if self.tmpmatrices is not None:
            L, R = self.tmpmatrices
            result = U.dot(L.dot(np.diag(S).dot(R.T.dot(xV.T))))
        else:
            result = U.dot(np.diag(S).dot(xV.T))
        
        if self.buffered:
            A, B = self.ab_buffer
            xB = x.T.dot(B)
            result += A.dot(xB.T)
            
        return result
    
    """ Compute the transposed matrix-vector M^T x  product without constructing the full
        matrix, where M is the compressed matrix and x is an input vector.
    """
    def trans_mat_vec_product(self, x):
        U,S,V = self.matrices
        
        # this term is written this way such that, if x is a sparse vector,
        # x's dot method is called. (Its clunky but its the only way I know)
        xU = x.T.dot(U)
        if self.tmpmatrices is not None:
            L, R = self.tmpmatrices
            result = V.dot(R.dot(np.diag(S).dot(L.T.dot(xU.T))))
        else:
            result = V.dot(np.diag(S).dot(xU.T))
        
        if self.buffered:
            A, B = self.ab_buffer
            xA = x.T.dot(A)
            result += B.dot(xA.T)
            
        return result
    
    """ Compute the inversed matrix-vector inv(M)x product, ignoring any deferred
        updates, where M is the compressed matrix and x is an input vector.
        Currently, there is no regularization of the inverse.
    """
    def inv_oldmat_vec_product(self, x, rcond = 1e-15, offset = 0.0):
        U,S,V = self.get_padded_non_zero_matrices(rcond)
        
        # this term is written this way such that, if x is a sparse vector,
        # x's dot method is called. (Its clunky but its the only way I know)
        xU = x.T.dot(U)
        Sinv = np.zeros_like(S)
        Sinv[~mask] = 1.0/(S[~mask] + offset)
        if self.tmpmatrices is not None:
            L, R = self.tmpmatrices
            result = V.dot(R.dot(np.diag(Sinv).dot(L.T.dot(xU.T))))
        else:
            result = V.dot(np.diag(Sinv).dot(xU.T))
                    
        return result
    
    """ Compute the inversed matrix-vector inv(M)x product, where M is the 
        compressed matrix and x is an input vector. Note that this will
        trigger an svd update, if required.
        Currently, there is no regularization of the inverse.
    """
    def inv_mat_vec_product(self, x, rcond = 1e-15, offset = 0.0):
        
        # update SVD if currently using the buffers
        if self.buffered and self.count > 0:
            self.matrices = self.update_and_clear_buffer()
            
        U,S,V, mask = self.get_padded_non_zero_matrices(rcond)
        # this term is written this way such that, if x is a sparse vector,
        # x's dot method is called. (Its clunky but its the only way I know)
        xU = x.T.dot(U)
        Sinv = np.zeros_like(S)
        Sinv[~mask] = 1.0/(S[~mask] + offset)
        if self.tmpmatrices is not None:
            L, R = self.tmpmatrices
            result = V.dot(R.dot(np.diag(Sinv).dot(L.T.dot(xU.T))))
        else:
            result = V.dot(np.diag(Sinv).dot(xU.T))
                    
        return result
    
    
    """ Build a full, dense matrix of the approximated matrix. This forces
        an update if required.
    """
    def get_updated_full_matrix(self):
        # update SVD if currently using the buffers
        if self.buffered and self.count > 0:
            self.matrices = self.update_and_clear_buffer()
            
        U,S,V = self.matrices
        if self.tmpmatrices is not None:
            L, R = self.tmpmatrices
            U = U.dot(L)
            V = V.dot(R)
        
        
        return U.dot(np.diag(S).dot(V.T))
    
    """ Build a full, dense matrix of the approximated matrix with the
        unprocessed buffers added.
    """
    def get_full_matrix(self):
        U,S,V = self.matrices
    
        if self.tmpmatrices is not None:
            L, R = self.tmpmatrices
            U = U.dot(L)
            V = V.dot(R)        
        
        M = U.dot(np.diag(S).dot(V.T))
        # add buffers if currently using the buffers
        if self.buffered and self.count > 0:
            A, B = self.ab_buffer
            M += A.dot(B.T)
        return M
        
    def get_non_zero_matrices(self, rcond = 1e-15):
        U, S, V = self.matrices
        non_zero = (np.abs(S) > rcond*S[0]).nonzero()[0]
        return U[:, non_zero], S[non_zero], V[:, non_zero]
        
    def get_padded_non_zero_matrices(self, rcond):
        U, S, V = self.matrices
        S_pad = S.copy()
        non_zero = (np.abs(S) > rcond*S[0]).nonzero()[0]
    
        mask = np.ones(S.shape, dtype = np.bool)        
        mask[non_zero] = False
        S_pad[mask] = 0.0
        
        return U, S_pad, V, mask
        
