import numpy as np
import scipy.sparse
import scipy.sparse.linalg
import scipy.linalg

import time

from numpy.linalg import LinAlgError

from lrilstd.compressed import CompressedMatrix
from lrilstd.representation import RemoveAction

from itertools import izip

import __builtin__

try:
    __builtin__.profile
except AttributeError:
    # No line profiler, provide a pass-through version
    def profile(func): return func
    __builtin__.profile = profile

"""
authors: Clement Gehring
contact: gehring@csail.mit.edu
date: May 2015
"""

################## VARIOUS INTERFACES ########################################

class ValueFn(object):
    def __init__(self):
        pass
    
    def __call__(self, state):
        raise NotImplementedError("Subclasses should implement this!")

class IncrementalValueFn(ValueFn):
    def __init__(self):
        pass
    
    def update(self, s_t, r_t, s_tp1):
        raise NotImplementedError("Subclasses should implement this!")

class QValueFn(object):
    def __init__(self):
        pass
    
    """ Evaluate state-action pairs. Either state and action have the same
        number of rows or one of them has only one row (or is 1D). If state
        has one dimensions, then the query will evaluate the value of each
        action in that state. If only one actions is given, then the query
        will evaluate all values of executing that action in those states.
        
        state: state can be a row vector or a matrix of row vectors
        
        action: action can be a single actions or an array of actions.
    """
    def __call__(self, state, action):
        raise NotImplementedError("Subclasses should implement this!")


class IncrementalQValueFn(QValueFn):
    def __init__(self):
        pass
    
    """ Update function for transition sample. It is assumed that if 
        the episode starts s_t = None, and of the episode terminates
        s_tp1 = None.
        
        s_t: 1D array representing the state at the start of the transition
        
        a_t: the action at the start of the transition
        
        r_t: the reward received during this transition
        
        s_tp1: the state at the end of the transition
        
        a_tp1: the next action to be taken from the next state
    """
    def update(self, s_t, a_t, r_t, s_tp1, a_tp1):
        raise NotImplementedError("Subclasses should implement this!")


################## VARIOUS QVALUE FN IMPLEMENTATIONS ##########################
class tLSTDLambda(IncrementalQValueFn):
    """ Low rank incremental LSTD Lambda constructor.
    
        gamma: the discount factor
        
        lamb: the eligibility trace decay (lambda).
        
        phi: a state action projector
        
        max_rank: the maximum rank of the approximation
        
        deferred_updates:   boolean to specify whether SVD updates should be
                            batched or not.
        
        theta_update_interval: (optional) the rate at which theta (the value
                                function parameters) are updated. By default,
                                if updates are not deferred, they will be 
                                updated after every transition. This might be 
                                expensive. If default and updates are deferred,
                                theta will only be updated after each SVD
                                update.
    """
    def __init__(self,
                 gamma,
                 lamb, 
                 phi, 
                 max_rank, 
                 deferred_updates,
                 theta_update_interval = None,
                 rcond = 1e-15,
                 singular_value_offset = 0.0):
        
        self.phi = phi

        self.b = np.zeros(self.phi.size)
        self.cA = CompressedMatrix(max_rank = max_rank, 
                                  size = phi.size, 
                                  deferred_updates = deferred_updates)
        self.theta = np.zeros_like(self.b)
        
        self.gamma = gamma
        self.lamb = lamb
        self.z = np.zeros_like(self.b)
        
        self.count = 0
        
        self.update_interval = theta_update_interval
        
        # if SVD updates are deferred and no theta update interval
        # was not specified, update theta after each SVD update
        if deferred_updates and theta_update_interval is None:
            self.update_interval = max_rank
        
        # if SVD is not deferred and theta update interval was
        # not specified, update theta at every step
        if theta_update_interval is None:
            self.update_interval = 1
    
        self.rcond = rcond
        self.offset = singular_value_offset
        
    def __call__(self, state, action):
        phi_sa = self.phi(state, action)
        return phi_sa.dot(self.theta)
    @profile
    def update(self, s_t, a_t, r_t, s_tp1, a_tp1):
        # check is episode start
        if s_t is None:
            self.z[:] = 0.0
            return
        
        # check if state is terminal
        if s_tp1 is None:
            phi_t = self.phi(s_t, a_t)
            d = phi_t
        else:
            # get feature vector (batched for better performance)
            states = np.vstack((s_t, s_tp1))
            actions = np.vstack((a_t, a_tp1))
            
            phi_t, phi_tp1 = self.phi(states, actions)
            d = phi_t - self.gamma * phi_tp1
        
        if d.ndim == 1:
            d = d.reshape((1,-1))
        
        if d.shape[0] > 1:
            d = d.T
        
        
        # update traces
        z = self.gamma*self.lamb*self.z + phi_t
        
        # update b vector
        b = z * r_t + self.b
        
        # update A matrix (done in-place)
        self.cA.update_matrix(z.T, d.T)
        
        # if required, update theta (maybe forcing an SVD)
        self.count = (self.count + 1) % self.update_interval
        if self.count == 0:
            self.theta = np.array(self.cA.inv_mat_vec_product(b.T,
                                                              rcond = self.rcond,
                                                              offset = self.offset)).squeeze()
        
        # update saved vectors
        self.b = b
        self.z = z
    
    @property
    def A(self):
        return self.cA.get_full_matrix()
    
class ILSTDLambda(IncrementalQValueFn):
    def __init__(self):
        pass
    
    
class MiniBatchLSTDLambda(IncrementalQValueFn):
    """ Mini batched LSTD Lambda constructor.
    
        gamma: the discount factor
        
        lamb: the eligibility trace decay (lambda).
        
        phi: a state action projector
        
        theta_update_interval: (optional) the rate at which theta (the value
                                function parameters) are updated. By default,
                                they will be updated after every transition. 
                                This might be expensive. 
    """
    def __init__(self,
                 gamma,
                 lamb, 
                 phi,
                 theta_update_interval = None,
                 rcond = 1e-14):
        
        self.phi = phi
        
        self.b = np.zeros(self.phi.size)
        self.theta = np.zeros_like(self.b)
        self.A = np.zeros((self.phi.size, self.phi.size))
        
        self.gamma = gamma
        self.lamb = lamb
        self.z = None #np.zeros_like(self.b)
        
        self.count = 0
        
        # if update interval is negative, never update theta
        self.update_interval = theta_update_interval
        
        # if theta update interval was not specified, update theta 
        # at every step
        if theta_update_interval is None:
            self.update_interval = 1
            
        
    
        self.initialized = False
        self.rcond = rcond
    
    def __call__(self, state, action):
        phi_sa = self.phi(state, action)
        return phi_sa.dot(self.theta)
    
    def update(self, s_t, a_t, r_t, s_tp1, a_tp1):
        # check is episode start
        if s_t is None:
            self.z = None
            return
        
        # check if state is terminal
        if s_tp1 is None:
            phi_t = self.phi(s_t, a_t)
            d = phi_t
        else:
            # get feature vector (batched for better performance)
            states = np.vstack((s_t, s_tp1))
            actions = np.vstack((a_t, a_tp1))
            
            phi_t, phi_tp1 = self.phi(states, actions)
            d = phi_t - self.gamma * phi_tp1
        
        if d.ndim == 1:
            d = d.reshape((1,-1))
        
        if d.shape[0] > 1:
            d = d.T
        
        
        # update traces
        if self.z is not None:
            z = self.gamma*self.lamb*self.z + phi_t
            if scipy.sparse.issparse(z):
                z.data *= (np.abs(z.data)> 1e-5)
                z.eliminate_zeros()
        else:
            z = phi_t
            
        if z.ndim == 1:
            z = z.reshape((1,-1))
            
        # update b vector
        if self.initialized:
            b = z * r_t + self.b
        else:
            b = z * r_t

        # update A matrix
        # this is done this way so that A becomes a sparse matrix if present
        
        if self.initialized:
            A = z.T.dot(d) + self.A
        else:
            A = z.T.dot(d)
            self.initialized = True
        
        # if required, update theta (without forcing an SVD)
        if self.update_interval > 0:
            self.count = (self.count + 1) % self.update_interval
            if self.count == 0:
                self.update_theta(A, b)
            
        # update saved vectors and matrices
        self.b = b
        self.z = z
        self.A = A
        
    def update_theta(self, A, b):
        try:
            if scipy.sparse.issparse(A):
                self.theta = scipy.sparse.linalg.lsmr(A,  
                                                      b.toarray().squeeze())[0]
            else:
                self.theta = np.linalg.lstsq(A, b.squeeze(), rcond = self.rcond)[0]
        except LinAlgError as e:
            print 'Least-squares failed...', e

        
class TDLambda(IncrementalQValueFn):
    def __init__(self, 
                 alpha, 
                 gamma, 
                 lamb,
                 phi, 
                 replacing_trace = True):
        self.gamma = gamma
        self.alpha = alpha
        self.lamb = lamb
        self.phi = phi
        self.theta = np.zeros(phi.size)
        self.replacing_trace = replacing_trace
        self.e = None
        
    def __call__(self, state, action):
        phi_sa = self.phi(state, action)
        return phi_sa.dot(self.theta)
    
    def update(self, s_t, a_t, r_t, s_tp1, a_tp1):
        # check is episode start
        if s_t is None:
            self.e = None
            return
        
        # check if state is terminal
        if s_tp1 is None:
            phi_t = self.phi(s_t, a_t)
            delta = r_t - phi_t.dot(self.theta)
        else:
            # get feature vector (batched for better performance)
            states = np.vstack((s_t, s_tp1))
            actions = np.vstack((a_t, a_tp1))
            
            phi_t, phi_tp1 = self.phi(states, actions)
            delta = r_t + self.gamma * phi_tp1.dot(self.theta) \
                            - phi_t.dot(self.theta)
        
        # update traces
        if self.e is not None:
            self.e = self.gamma*self.lamb*self.e + phi_t
            if scipy.sparse.issparse(self.e):
                self.e.data *= (np.abs(self.e.data)> 1e-5)
                self.e.eliminate_zeros()
            # clip traces if set to replace
                if self.replacing_trace:
                    np.clip(self.e.data, 0.0, 1.0, out = self.e.data)
            elif self.replacing_trace:
                np.clip(self.e, 0.0, 1.0, out = self.e)
        else:
            self.e = phi_t
        
        # update weights
        self.theta += (self.alpha * delta * self.e)
        self.theta = np.array(self.theta).squeeze()
        
################## VARIOUS HELPER FUNCTIONS ##################################

""" Batch form of LSTD. This should be significantly faster than the iterative
    version implemented as a class given that enough memory is available.
"""
def batch_LSTD(trajectories, 
               rewards, 
               phi, 
               gamma,
               lamb = 0.0,
               rcond = 1e-14,
               minibatch_size = None,
               max_num_samples = None,
               timed = False, 
               no_sparse = False, 
               solve = True,
               **args):
                   
    if timed:
        start_time = time.clock()
    if lamb == 0.0:
        
        phi_t, phi_tp1, R = make_phi_R_matrix(trajectories, rewards, phi, max_num_samples)
        if minibatch_size is None:
            indices = np.array([0, phi_t.shape[0]], dtype='int')
        else:
            indices = np.hstack((np.arange(0, phi_t.shape[0], minibatch_size), 
                                 np.array([phi_t.shape[0]], dtype='int')))
            
            
            
        b = np.zeros(phi_t.shape[1])
        if scipy.sparse.isspmatrix(phi_t):
            A = scipy.sparse.csr_matrix((phi_t.shape[1], phi_t.shape[1]))
        else:
            A = np.zeros((phi_t.shape[1], phi_t.shape[1]))
            
        count = 0
        for i in xrange(indices.size-1):
            i0, i1 = indices[i], indices[i+1]
            count += i1- i0
            A += phi_t[i0:i1, :].T.dot(phi_t[i0:i1, :] - phi_tp1[i0:i1, :]*gamma)
            b += phi_t[i0:i1, :].T.dot(R[i0:i1])
            
            # solve for parameters
            if solve:
                if scipy.sparse.isspmatrix(A):
                    theta = scipy.sparse.linalg.lsmr(A,b)[0]
                else:
                    theta = np.linalg.lstsq(A, b, rcond=rcond)[0]
    else:
        lstd = MiniBatchLSTDLambda(gamma, lamb, RemoveAction(phi), theta_update_interval = -1, rcond = rcond)
        count = 0
        a = np.ones(1)
        done = False
            
        for traj, rew in izip(trajectories, rewards):
            s_t = traj[0]
            for s_tp1, r_t in izip(traj[1:], rew):
                count += 1
                lstd.update(s_t, a, r_t, s_tp1, a)
                s_t = s_tp1
                if max_num_samples is not None and count >= max_num_samples:
                    done = True
                    break
            if done:
                break
            
            A = lstd.A
            b = lstd.b
            lstd.update_theta(A, b)
            theta = lstd.theta
                        
    
    if timed:
        runtime = time.clock() - start_time
    
    if timed:
        output = (A/count, b/count, theta, runtime)
    elif solve:
        output = (A/count, b/count, theta)
    else:
        output = (A/count, b/count)
    return output

""" batch version of tLSTD
"""
def batch_tLSTD(trajectories,
                  rewards,
                  phi,
                  gamma,
                  max_rank,
                  lamb = 0.0,
                  rcond = 1e-14,
                  singular_value_offset = 0.0,
                  minibatch_size = None,
                  max_num_samples = None,
                  timed = False,
                  **args):
    if timed:
        start_time = time.clock()
    if lamb == 0.0:
        A = CompressedMatrix(max_rank, phi.size, deferred_updates = True)
        phi_t, phi_tp1, R = make_phi_R_matrix(trajectories, rewards, phi, max_num_samples)
        if minibatch_size is None:
            indices = np.array([0, phi_t.shape[0]], dtype='int')
        else:
            indices = np.hstack((np.arange(0, phi_t.shape[0], minibatch_size), 
                                 np.array([phi_t.shape[0]], dtype='int')))
            
            
            
        b = np.zeros(phi_t.shape[1])
        
            
        for i in xrange(len(indices)-1):
            i0, i1 = indices[i], indices[i+1]
            A.update_matrix(phi_t[i0:i1, :].T, (phi_t[i0:i1, :] - phi_tp1[i0:i1, :]*gamma).T)
            b += phi_t[i0:i1, :].T.dot(R[i0:i1])
        theta = A.inv_mat_vec_product(b, rcond, singular_value_offset) 
        
    else:
        lrilstd = tLSTDLambda(gamma, lamb, RemoveAction(phi), max_rank, deferred_updates = True, theta_update_interval = minibatch_size, rcond = rcond, singular_value_offset = 0.0)
        count = 0
        a = np.ones(1)
        done = False
            
        for traj, rew in izip(trajectories, rewards):
            s_t = traj[0]
            for s_tp1, r_t in izip(traj[1:], rew):
                count += 1
                lrilstd.update(s_t, a, r_t, s_tp1, a)
                s_t = s_tp1
                if max_num_samples is not None and count >= max_num_samples:
                    done = True
                    break
            if done:
                break
            
            A = lrilstd.A
            b = lrilstd.b
            theta = lrilstd.theta
    
    if timed:
        runtime = time.clock() - start_time
    
    if timed:
        output = (A, b, theta, runtime)
    else:
        output = (A, b, theta)
    return output
    
def make_phi_R_matrix(trajectories,
                      rewards,
                      phi, 
                      max_num_samples = None):
    # apply the phi projection while keeping track of which states were 
    # terminal
    all_phi = []
    num_samples = 0
    for states in trajectories:
        
        if states[-1] is None:
            state_array = np.vstack(states[:-1])
            phis = phi(state_array)
            if scipy.sparse.isspmatrix(phis):
                phis = scipy.sparse.vstack((phis, scipy.sparse.csr_matrix((1, phis.shape[1]))), 'csr')
            else:
                if phis.ndim == 1:
                    phis = phis.reshape((1,-1))
                phis = np.concatenate((phis, np.zeros((1,phis.shape[1]))), axis=0)
        else:
            phis = phi(np.concatenate(states, axis=0))
    
        num_samples += phis.shape[0]-1
        phi_t = phis[:-1,:]
        phi_tp1 = phis[1:, :]
        all_phi.append((phi_t, phi_tp1))
        if max_num_samples is not None and num_samples >= max_num_samples:
            break
        
    # build phi_t and phi_tp1 for all trajectories
    if scipy.sparse.isspmatrix(all_phi[0][0]):
        x_t, x_tp1 = zip(*all_phi)
        phi_t = scipy.sparse.vstack(x_t, 'csc')
        phi_tp1 = scipy.sparse.vstack(x_tp1, 'csc')
    else:
        x_t, x_tp1 = zip(*all_phi)
        phi_t = np.vstack(x_t)
        phi_tp1 = np.vstack(x_tp1)
        
    # concatenate all the rewards    
    R = np.hstack(rewards).squeeze()
    
    if max_num_samples is not None:
        end_index = min(max_num_samples, phi_t.shape[0])
        phi_t = phi_t[:end_index, :]
        phi_tp1 = phi_tp1[:end_index,:]
        if scipy.sparse.issparse(phi_t):
            phi_t = phi_t.tocsr()
            phi_tp1 = phi_tp1.tocsr()
        R = R[:end_index]
        
    return phi_t, phi_tp1, R

""" Single pass batch version for linear TD(lambda), this is not faster than the
    regular td implementation and is included for convenience only.
"""
def batch_linear_TD(trajectories,
                    rewards,
                    phi,
                    alpha,
                    gamma,
                    lamb,
                    replacing_trace = True,
                    max_num_samples = None,
                    timed = False,
                    **args):
    phi = RemoveAction(phi)
    td_val = TDLambda(alpha, gamma, lamb, phi, replacing_trace)
    count = 0
    a = np.ones(1)
    done = False
    
    if timed:
        start_time = time.clock()
        
    for traj, rew in izip(trajectories, rewards):
        s_t = traj[0]
        for s_tp1, r_t in izip(traj[1:], rew):
            count += 1
            td_val.update(s_t, a, r_t, s_tp1, a)
            s_t = s_tp1
            if max_num_samples is not None and count >= max_num_samples:
                done = True
                break
        if done:
            break
        
    if timed:
        runtime = time.clock() - start_time
        
    if timed:
        output = (td_val.theta, runtime)
    else:
        output = (td_val.theta, )
    return output
        
